package com.dsk.flink.dsc.common.function;

import cn.hutool.core.collection.CollUtil;
import cn.hutool.core.map.MapUtil;
import cn.hutool.core.util.StrUtil;
import com.alibaba.fastjson.JSONArray;
import com.alibaba.fastjson.JSONObject;
import com.dsk.flink.dsc.utils.EnvProperties;
import org.apache.flink.api.java.tuple.Tuple3;
import org.apache.flink.configuration.Configuration;
import org.apache.flink.streaming.api.functions.async.ResultFuture;
import org.apache.flink.streaming.api.functions.async.RichAsyncFunction;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.text.SimpleDateFormat;
import java.util.*;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.LinkedBlockingDeque;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;

public class AsyncMysqlDataTransferFunctionNew extends RichAsyncFunction<JSONObject, Tuple3<String,String,Long>> {

    //static Logger logger = LoggerFactory.getLogger(AsyncMysqlDataTransferFunctionNew.class);
    //数据库连接信息
    EnvProperties dbInfoMap;

    //线程池
    private transient ExecutorService executorService;

    public AsyncMysqlDataTransferFunctionNew(EnvProperties dbInfoMap) {
        this.dbInfoMap = dbInfoMap;
    }

    @Override
    public void open(Configuration parameters) throws Exception {
        //初始化线程池
        executorService = new ThreadPoolExecutor(12 , 12, 20, TimeUnit.MINUTES, new LinkedBlockingDeque<>());
    }

    @Override
    public void close() throws Exception {
        executorService.shutdown();
    }

    @Override
    public void timeout(JSONObject input, ResultFuture<Tuple3<String,String,Long>> resultFuture) throws Exception {
        resultFuture.complete(Collections.singleton(Tuple3.of("err","",0L)));
    }

    @Override
    public void asyncInvoke(JSONObject value, ResultFuture<Tuple3<String,String,Long>> resultFuture) throws Exception {
        executorService.submit(() -> {
            try {
                //返回数据集合
                List<Tuple3<String,String,Long>> resultList = new ArrayList<>();
                String type = value.getString("type");
                JSONArray dataList = value.getJSONArray("data");
                JSONObject mysqlType = value.getJSONObject("mysqlType");
                //JSONArray oldDataList = value.getJSONArray("old");
                String table = value.getString("table");
                Boolean isDdl = value.getBoolean("isDdl");
                JSONArray pkNames = value.getJSONArray("pkNames");
                Set<String> pkNameSet = new HashSet<>();
                long ts = value.getLong("ts") == null ? System.currentTimeMillis() : value.getLong("ts");
                if(CollUtil.isNotEmpty(pkNames)){
                    pkNames.forEach(name -> pkNameSet.add(String.valueOf(name)));
                }

                String excueteSql = "";
                if(isDdl){
                    excueteSql = value.getString("sql");
                    if(StrUtil.isNotBlank(excueteSql)){
                        excueteSql = StrUtil.subBefore(excueteSql,"AFTER",true);
                        resultList.add(Tuple3.of(excueteSql,table,ts));
                    }
                    resultFuture.complete(resultList);
                    return;
                }
                JSONObject dataObj = dataList.getJSONObject(0);

                Boolean logicalDelete = MapUtil.getBool(dbInfoMap, "logical_delete", false);
                if(logicalDelete){
                    mysqlType.put("is_del", "int");
                    dataObj.put("is_del", "DELETE".equals(type) ? 1 : 0);
                }

                //处理先后顺序
                //获取该条数据的表名和主键作为唯一的groupKey
                String groupKey = table;
                for (String pk : pkNameSet) {
                    String pkValue = getValueString(dataObj, pk, mysqlType.getString(pk));
                    groupKey = table.concat("-").concat(pkValue);
                }

                if("INSERT".equals(type)){
                    excueteSql = tranferInsertSql(table,dataObj,mysqlType);
                }
                if("UPDATE".equals(type)){
                    //JSONObject oldDataObj = oldDataList.getJSONObject(0);
                    //excueteSql = tranferUpdateSql(table,dataObj,oldDataObj,mysqlType,pkNameSet);
                    excueteSql = tranferInsertSql(table,dataObj,mysqlType);
                }

                if("DELETE".equals(type)){
                    excueteSql = logicalDelete ? tranferInsertSql(table,dataObj,mysqlType) :  transferDeleteSql(table,dataObj,mysqlType,pkNameSet);
                }
                resultList.add(Tuple3.of(excueteSql,groupKey,ts));
                Boolean logEnable = MapUtil.getBool(dbInfoMap, "log_enable", false);
                if (logEnable){
                    //logger.error("-----datajson:{}----",value.toJSONString());
                    String logSql = buildLogData(type, table, pkNameSet, dataObj, ts, value.toJSONString());
                    resultList.add(Tuple3.of(logSql,"dsc_cdc_log",ts));
                }
                resultFuture.complete(resultList);

            }catch (Exception e){
                e.printStackTrace();
                resultFuture.complete(Collections.singleton(Tuple3.of("err","",0L)));
            }
        });
    }

    private static String logSqlFormat = "INSERT INTO dsc_cdc_log (`table`,op_type,pk_columns,pk_values,data_json,cdc_ts) values ('%s','%s','%s','%s','%s', %d)";

    private String buildLogData(String type, String table, Set<String> pkNameSet, JSONObject dataObj, long ts, String dataJsonStr) {
        List<String> pkValueList = new ArrayList<>();
        for (String pk : pkNameSet) {
            pkValueList.add(dataObj.getString(pk));
        }
        String pkColumns = String.join(",",pkNameSet);
        String pkValues = String.join("-",pkValueList);
        dataJsonStr = dataJsonStr.replace("\\","\\\\");
        return String.format(logSqlFormat, table, type, pkColumns, pkValues, dataJsonStr, ts);
    }


    private static final String[] STR_SQL_TYPE = new String[]{"VARCHAR","CHAR","TINYBLOB","BLOB","MEDIUMBLOB","LONGBLOB","TINYTEXT","TEXT","MEDIUMTEXT","LONGTEXT","TIME","TIMESTAMP","JSON","json"};

    private static final String[] KEYWORD = new String[]{"limit"};
    private static String tranferInsertSql(String table, JSONObject dataObj, JSONObject mysqlType) {

        Set<String> columnSet = mysqlType.keySet();
        StringBuilder sb = new StringBuilder();
        for (String s : columnSet) {
            sb.append("`"+s+"`").append(",");
        }
        List<String> valueList = new ArrayList<>();
        for (String col : columnSet) {
            if(Arrays.asList(KEYWORD).contains(col)){
                valueList.add(getValueString(dataObj,col,mysqlType.getString(col)));
            }else {
                valueList.add(getValueString(dataObj,col,mysqlType.getString(col)));
            }
        }
        sb.setLength(sb.length()-1);
        String columnString = sb.toString();
        String valueString = String.join(",",valueList);

        //return String.format("INSERT INTO %s (%s) values (%s) ON DUPLICATE KEY UPDATE %s;",table,columnString,valueString,updateString);
        return String.format("REPLACE INTO %s (%s) values (%s);",table,columnString,valueString);
    }

    private String tranferUpdateSql(String table, JSONObject dataObj, JSONObject oldDataObj, JSONObject mysqlType,Set<String> pkNameSet) {

        Set<String> columnSet = mysqlType.keySet();
        List<String> setList = new ArrayList<>();
        List<String> whereList = new ArrayList<>();
        for (String col : columnSet) {
            String setString = col.concat(" = ").concat(getValueString(dataObj,col,mysqlType.getString(col)));
            setList.add(setString);
        }

        for (String pk : pkNameSet) {
            String whereString = pk.concat(" = ").concat(getValueString(oldDataObj,pk,mysqlType.getString(pk)));
            whereList.add(whereString);
        }
        String setString = String.join(",",setList);
        String whereString = String.join(" and ",whereList);

        return String.format("UPDATE %s SET %s WHERE %s",table,setString,whereString);
    }

    private String transferDeleteSql(String table, JSONObject dataObj, JSONObject mysqlType, Set<String> pkNameSet) {

        List<String> whereList = new ArrayList<>();
        for (String pk : pkNameSet) {
            String whereString = pk.concat(" = ").concat(getValueString(dataObj,pk,mysqlType.getString(pk)));
            whereList.add(whereString);
        }
        String whereString = String.join(" and ",whereList);
        return String.format("DELETE FROM %s WHERE %s",table,whereString);
    }

    /**
     * @author shezaixing
     * @date 2023/12/7 14:23
     * @description 判断拼接字符串时类型（是否需要加上引号）
     *
     */
    private static String getValueString(JSONObject dataObj,String columnKey,String mysqlType){

        if(null == dataObj.get(columnKey)){
            return "null";
        }

        //需要处理成字符串加引号的类型
        if(Arrays.asList(STR_SQL_TYPE).contains(mysqlType.toUpperCase())){
            return String.format("'%s'", dataObj.getString(columnKey).replace("\\","\\\\").replace("'", "\\'") );
        }

        //时间字段处理
        if("DATE".equalsIgnoreCase(mysqlType) || "DATETIME".equalsIgnoreCase(mysqlType)){
            SimpleDateFormat df = "DATETIME".equalsIgnoreCase(mysqlType) ? new SimpleDateFormat("yyyy-MM-dd HH:mm:ss") : new SimpleDateFormat("yyyy-MM-dd");
            return String.format("\"%s\"",df.format(dataObj.getDate(columnKey)));
        }

        return dataObj.getString(columnKey);
    }

}
